unit CoreExports;

interface

uses
  Windows;

type
  TDACEntry = packed record
    Channel: Word;
    Value: Word;
  end;
  TPDACEntry = ^TDACEntry;

procedure InitDLL;

function HWave_Start(DeviceIndex: LongWord; Milliseconds: Double; BufferDepth: LongInt): LongWord; cdecl;
function HWave_Stop(DeviceIndex: LongWord): LongWord; cdecl;
function HWave_Queue(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
function HWave_QueueMulti(DeviceIndex: LongWord; pDACData: TPDACEntry; DACDataCount: LongWord): LongWord; cdecl;
function HWave_FlushQueue(DeviceIndex: LongWord): LongWord; cdecl;
function HWave_ResetClock(DeviceIndex: LongWord): LongWord; cdecl;
function HWave_ChangeClockRate(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
function HWave_QueryClocks(DeviceIndex: LongWord; pClocks: PDouble; ClockCount: LongWord): LongWord; cdecl;



implementation

uses
  Math,
  SysUtils, Classes,
  AIOUSB;

const
  AUR_DAC_IMMEDIATE       = $B3;

type
  TWavePoint = packed record
    AFlags: Byte;
    AData: array[0..7] of Word;
    BFlags: Byte;
    BData: array[0..7] of Word;
  end;

  THWaveThread = class(TThread)
  public
    TicksPerSec, TicksPerPoint, NextPoint: Int64;
    BufferDepth: LongInt;
    WavePointQueue: array of TWavePoint;
    iWavePointRead, iWavePointWrite, WavePointDepthLeft: Integer;
    hWavePointMutex, hWavePointSem: THandle;
    bStarted: Boolean;
    NewTicksPerPoint: Int64;
    bHasNewTicksPerPoint: Boolean;

    constructor Create(NewDeviceIndex: LongWord; NewMilliseconds: Double; NewBufferDepth: LongInt);
    destructor Destroy; override;
  protected
    DeviceIndex: LongWord;

    procedure Execute; override;
  end;

var
  HWaveThread: array[0..31] of THWaveThread;
  hThreadTableMutex: THandle;

procedure DLLHandler(Reason: Integer);
var
  I: Integer;

begin
  if (Reason = DLL_PROCESS_DETACH) then begin
    WaitForSingleObject(hThreadTableMutex, INFINITE);
      for I := 0 to High(HWaveThread) do
        FreeAndNil(HWaveThread[I])
      ;
    ReleaseMutex(hThreadTableMutex);
    CloseHandle(hThreadTableMutex);
    hThreadTableMutex := INVALID_HANDLE_VALUE;
  end;
end;

procedure InitDLL;
begin
  if PPointer(@DLLProc)^ = @DLLHandler then Exit;

  hThreadTableMutex := CreateMutex(nil, False, nil);
  DLLProc := @DLLHandler;
end;

function ValidateDI(var DI: LongWord): LongWord;
begin
  if DI < LongWord(Length(hWaveThread)) then begin
    Result := ERROR_SUCCESS;
    Exit;
  end;

  DI := ResolveDeviceIndex(DI);
  if DI >= LongWord(Length(hWaveThread)) then begin
    Result := GetLastError;
    if Result = ERROR_SUCCESS then Result := ERROR_INVALID_INDEX;
    Exit;
  end;

  Result := ERROR_SUCCESS;
end;

function HWave_Start(DeviceIndex: LongWord; Milliseconds: Double; BufferDepth: LongInt): LongWord; cdecl;
var
  NewThread: THWaveThread;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if (Milliseconds <= 0) or (BufferDepth <= 0) then begin
      Result := ERROR_INVALID_PARAMETER;
      Exit;
    end;

    NewThread := THWaveThread.Create(DeviceIndex, Milliseconds, BufferDepth);

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] <> nil then begin
        Result := ERROR_SINGLE_INSTANCE_APP;
      end
      else begin
        Result := ERROR_SUCCESS;
        HWaveThread[DeviceIndex] := NewThread;
      end;
    ReleaseMutex(hThreadTableMutex);

    if Result = ERROR_SUCCESS then
      NewThread.Start
    else
      NewThread.Free
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_Stop(DeviceIndex: LongWord): LongWord; cdecl;
var
  OldThread: THWaveThread;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    OldThread := nil;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        Result := ERROR_SUCCESS;
        OldThread := HWaveThread[DeviceIndex];
        HWaveThread[DeviceIndex] := nil;
      end;
    ReleaseMutex(hThreadTableMutex);

    if OldThread <> nil then
      OldThread.Free
    ;
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_QueueWavePoint(DeviceIndex: LongWord; const NewWavePoint: TWavePoint): LongWord;
begin
  try
    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        WaitForSingleObject(HWaveThread[DeviceIndex].hWavePointMutex, INFINITE);
          if ReleaseSemaphore(HWaveThread[DeviceIndex].hWavePointSem, 1, nil) then begin
            //It's safe to release the semaphore first because we hold the mutex.
            Result := ERROR_SUCCESS;
            HWaveThread[DeviceIndex].WavePointQueue[HWaveThread[DeviceIndex].iWavePointWrite] := NewWavePoint;
            HWaveThread[DeviceIndex].iWavePointWrite := (HWaveThread[DeviceIndex].iWavePointWrite + 1) mod HWaveThread[DeviceIndex].BufferDepth;
          end
          else begin
            Result := ERROR_NOT_READY;
          end;
        ReleaseMutex(HWaveThread[DeviceIndex].hWavePointMutex);
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_Queue(DeviceIndex: LongWord; pData: PWord): LongWord; cdecl;
var
  NewWavePoint: TWavePoint;
  iDAC: Integer;
  pNextData: PWord;
  ProbeStr: AnsiString;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    try
      SetString(ProbeStr, PAnsiChar(pData), 16*SizeOf(pData^));
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;

    pNextData := pData;
    NewWavePoint.AFlags := $FF;
    NewWavePoint.BFlags := $FF;
    for iDAC := 0 to 7 do begin
      NewWavePoint.AData[iDAC] := pNextData^;
      Inc(pNextData);
    end;
    for iDAC := 0 to 7 do begin
      NewWavePoint.BData[iDAC] := pNextData^;
      Inc(pNextData);
    end;

    Result := HWave_QueueWavePoint(DeviceIndex, NewWavePoint);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_QueueMulti(DeviceIndex: LongWord; pDACData: TPDACEntry; DACDataCount: LongWord): LongWord; cdecl;
const
  ImmDACs = 16;

var
  NewWavePoint: TWavePoint;
  ProbeStr: AnsiString;
  pDE: TPDACEntry;
  pDB, pMB: PByte;
  I, J: Integer;
  L: LongWord;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    if DACDataCount = 0 then Exit; //If length is zero, then it's a nop.

    try
      SetString(ProbeStr, PAnsiChar(pDACData), DACDataCount*SizeOf(pDACData^));
    except
      Result := ERROR_NOACCESS;
      Exit;
    end;



    pDE := pDACData;
    L := 0;
    for I := 0 to DACDataCount - 1 do begin
      if pDE.Channel > L then L := pDE.Channel;
      Inc(pDE);
    end;
    //L is now highest channel.

    if L >= ImmDACs then begin
      Result := ERROR_INVALID_ADDRESS;
      Exit;
    end;

    ZeroMemory(@NewWavePoint, SizeOf(NewWavePoint));
    pDB := PByte(@NewWavePoint);
    pDE := pDACData;
    for I := 0 to DACDataCount - 1 do begin
      J := pDE.Channel;
      pMB := pDB;
      Inc(pMB, J * 2 + (J div 8) + 1);
      PWord(pMB)^ := pDE.Value;

      pMB := pDB;
      Inc(pMB, (J div 8) * 17);
      pMB^ := pMB^ or (1 shl (J and 7));

      Inc(pDE);
    end;

    Result := HWave_QueueWavePoint(DeviceIndex, NewWavePoint);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_FlushQueue(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        WaitForSingleObject(HWaveThread[DeviceIndex].hWavePointMutex, INFINITE);
          while WaitForSingleObject(HWaveThread[DeviceIndex].hWavePointSem, 0) = WAIT_OBJECT_0 do begin
            with HWaveThread[DeviceIndex] do begin
              Dec(iWavePointWrite);
              if iWavePointWrite < 0 then iWavePointWrite := iWavePointWrite + BufferDepth;
            end;
          end;
        ReleaseMutex(HWaveThread[DeviceIndex].hWavePointMutex);
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_ResetClock(DeviceIndex: LongWord): LongWord; cdecl;
begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        HWaveThread[DeviceIndex].bStarted := False;
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_ChangeClockRate(DeviceIndex: LongWord; Milliseconds: Double): LongWord; cdecl;
var
  NewTicksPerPoint: Int64;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then Exit;

    QueryPerformanceFrequency(NewTicksPerPoint);
    NewTicksPerPoint := Round(NewTicksPerPoint * Milliseconds / 1000);

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        HWaveThread[DeviceIndex].NewTicksPerPoint := NewTicksPerPoint;
        HWaveThread[DeviceIndex].bHasNewTicksPerPoint := True;
      end;
    ReleaseMutex(hThreadTableMutex);
  except
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

function HWave_QueryClocks(DeviceIndex: LongWord; pClocks: PDouble; ClockCount: LongWord): LongWord; cdecl;
const
  NAN = 0/0;

var
  iClock: Integer;
  TicksPerSec: Int64;
  MillisecsPerTick: Extended;
  ClockTicks: array of Int64;
  pThisClock: PDouble;

  procedure BlankClocks;
  var
    iClock: Integer;
    pThisClock: PDouble;

  begin
    try
      pThisClock := pClocks;
      for iClock := 0 to ClockCount-1 do begin
        pThisClock^ := NAN;
        Inc(pThisClock);
      end;
    except
      if Result = ERROR_SUCCESS then Result := ERROR_NOACCESS;
    end;
  end;

begin
  try
    Result := ValidateDI(DeviceIndex);
    if Result <> ERROR_SUCCESS then begin
      BlankClocks;
      Exit;
    end;

    SetLength(ClockTicks, ClockCount);

    WaitForSingleObject(hThreadTableMutex, INFINITE);
      if HWaveThread[DeviceIndex] = nil then begin
        Result := ERROR_SERVICE_NOT_ACTIVE;
      end
      else begin
        if ClockCount > 0 then ClockTicks[0] := HWaveThread[DeviceIndex].TicksPerPoint;
        if ClockCount > 1 then ClockTicks[1] := HWaveThread[DeviceIndex].NextPoint;
        if ClockCount > 2 then QueryPerformanceCounter(ClockTicks[2]);
      end;
    ReleaseMutex(hThreadTableMutex);

    BlankClocks;

    if Result <> ERROR_SUCCESS then Exit;

    QueryPerformanceFrequency(TicksPerSec);
    MillisecsPerTick := 1000/TicksPerSec;
    pThisClock := pClocks;
    for iClock := 0 to Min(2, ClockCount-1) do begin
      pThisClock^ := MillisecsPerTick * ClockTicks[iClock];
      Inc(pThisClock);
    end;

  except
    BlankClocks;
    Result := ERROR_INTERNAL_ERROR;
  end;
end;

{ THWaveThread }

constructor THWaveThread.Create(NewDeviceIndex: LongWord; NewMilliseconds: Double; NewBufferDepth: LongInt);
begin
  inherited Create(True);

  DeviceIndex := NewDeviceIndex;
  BufferDepth := NewBufferDepth;

  hWavePointMutex := CreateMutex(nil, False, nil);
  hWavePointSem := CreateSemaphore(nil, 0, BufferDepth-1, nil);
  SetLength(WavePointQueue, BufferDepth);

  QueryPerformanceFrequency(TicksPerSec);
  TicksPerPoint := Round(TicksPerSec * NewMilliseconds / 1000);
end;

destructor THWaveThread.Destroy;
begin
  Terminate;
  ReleaseSemaphore(hWavePointSem, 1, nil); //This is a "kill release".

  inherited;

  CloseHandle(hWavePointMutex);
  CloseHandle(hWavePointSem);
end;

procedure THWaveThread.Execute;
var
  Naui: Int64;
  ThisWavePoint: TWavePoint;
  Text, Caption: String;

begin
  try
    bStarted := False;
    SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_TIME_CRITICAL);
    repeat
      WaitForSingleObject(hWavePointSem, INFINITE);
      if Terminated then Break;

      WaitForSingleObject(hWavePointMutex, INFINITE);
        ThisWavePoint := WavePointQueue[iWavePointRead];
      ReleaseMutex(hWavePointMutex);
      iWavePointRead := (iWavePointRead + 1) mod BufferDepth;

      if bStarted then begin
        repeat
          QueryPerformanceCounter(Naui);
          if (Naui - NextPoint) > 0 then Break;
          if Terminated then Break;

          if bHasNewTicksPerPoint then begin
            bHasNewTicksPerPoint := False;
            TicksPerPoint := NewTicksPerPoint;
          end;
        until False;
        if Terminated then Break;
      end
      else begin
        QueryPerformanceCounter(NextPoint);
        bStarted := True;
      end;
      Inc(NextPoint, TicksPerPoint);

      GenericVendorWrite(DeviceIndex, AUR_DAC_IMMEDIATE, 0, 0, SizeOf(ThisWavePoint), @ThisWavePoint);
    until Terminated;
  except
    on E: Exception do begin
      SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_NORMAL);

      Caption := 'HWave';
      Text := E.ClassName + 'exception from alternate thread context:'#13#10 + E.Message;
      MessageBox(0, PChar(Text), PChar(Caption), MB_ICONEXCLAMATION);
      Text := '';
      Caption := '';
    end;
  end;
end;

end.
